@@ -14,6 +14,7 @@
 
 
 import hashlib
+import json
 import logging
 from datetime import datetime, timezone
 
@@ -35,7 +36,8 @@ def test_verifier_production():
     assert verifier is not None
 
 
-def test_verifier_staging(mock_staging_tuf):
+@pytest.mark.staging
+def test_verifier_staging():
     verifier = Verifier.staging()
     assert verifier is not None
 
@@ -50,7 +52,7 @@ def test_verifier_one_verification(signing_materials, null_policy):
 
 
 @pytest.mark.staging
-def test_verifier_inconsistent_log_entry(signing_bundle, null_policy, mock_staging_tuf):
+def test_verifier_inconsistent_log_entry(signing_bundle, null_policy):
     (file, bundle) = signing_bundle("bundle_cve_2022_36056.txt")
 
     verifier = Verifier.staging()
@@ -75,15 +77,28 @@ def test_verifier_multiple_verifications(signing_materials, null_policy):
 
 @pytest.mark.online
 @pytest.mark.parametrize(
-    "filename", ("bundle.txt", "bundle_v3.txt", "bundle_v3_alt.txt")
+    "filename",
+    ("bundle.txt", "bundle_v3.txt", "bundle_v3_alt.txt", "staging-rekor-v2.txt"),
 )
-def test_verifier_bundle(signing_bundle, null_policy, filename):
+def test_verifier_bundle_artifact(signing_bundle, null_policy, filename):
     (file, bundle) = signing_bundle(filename)
 
     verifier = Verifier.staging()
     verifier.verify_artifact(file.read_bytes(), bundle, null_policy)
 
 
+@pytest.mark.online
+@pytest.mark.parametrize(
+    "filename",
+    ("a.dsse.staging-rekor-v2.txt",),
+)
+def test_verifier_bundle_dsse(signing_bundle, null_policy, filename):
+    (file, bundle) = signing_bundle(filename)
+
+    verifier = Verifier.staging()
+    verifier.verify_dsse(bundle, null_policy)
+
+
 @pytest.mark.parametrize(
     "filename", ("bundle.txt", "bundle_v3.txt", "bundle_v3_alt.txt")
 )
@@ -216,6 +231,47 @@ class TestVerifierWithTimestamp:
             null_policy,
         )
 
+    def test_verifier_no_validity_end(self, verifier, asset, null_policy):
+        verifier._trusted_root.get_timestamp_authorities()[
+            0
+        ]._inner.valid_for.end = None
+        verifier.verify_artifact(
+            asset("tsa/bundle.txt").read_bytes(),
+            Bundle.from_json(asset("tsa/bundle.txt.sigstore").read_bytes()),
+            null_policy,
+        )
+
+    @pytest.mark.parametrize(
+        "fields_to_delete",
+        (
+            [],
+            ["inclusionPromise"],
+            # integratedTime is required to verify the inclusionPromise.
+            pytest.param(["integratedTime"], marks=pytest.mark.xfail),
+            ["inclusionPromise", "integratedTime"],
+        ),
+    )
+    def test_verifier_verify_no_inclusion_promise_and_integrated_time(
+        self, verifier, asset, null_policy, fields_to_delete
+    ):
+        """
+        Ensure that we can still verify a Bundle with an RFC 3161 timestamp if the SET isn't present.
+
+        There is one exception: When inclusionPromise is present, but integratedTime is not, then we expect a failure
+        because the integratedTime is required to verify the inclusionPromise.
+        """
+        bundle_dict = json.loads(asset("tsa/bundle.txt.sigstore").read_bytes())
+        (entry_dict,) = bundle_dict["verificationMaterial"]["tlogEntries"]
+        for field in fields_to_delete:
+            del entry_dict[field]
+        # Bundle.from_json() also validates the bundle's layout.
+        bundle = Bundle.from_json(json.dumps(bundle_dict))
+        verifier.verify_artifact(
+            asset("tsa/bundle.txt").read_bytes(),
+            bundle,
+            null_policy,
+        )
+
     def test_verifier_without_timestamp(
         self, verifier, asset, null_policy, monkeypatch
     ):
@@ -245,32 +301,6 @@ class TestVerifierWithTimestamp:
                 null_policy,
             )
 
-    def test_verifier_no_validity(
-        self, caplog, verifier, asset, null_policy, monkeypatch
-    ):
-        # asset is a rekor v1 bundle: set threshold to 2 so both integrated time and the
-        # TSA timestamp are required
-        monkeypatch.setattr("sigstore.verify.verifier.VERIFIED_TIME_THRESHOLD", 2)
-
-        verifier._trusted_root.get_timestamp_authorities()[
-            0
-        ]._inner.valid_for.end = None
-
-        with caplog.at_level(logging.DEBUG, logger="sigstore.verify.verifier"):
-            with pytest.raises(
-                VerificationError, match="not enough sources of verified time"
-            ):
-                verifier.verify_artifact(
-                    asset("tsa/bundle.txt").read_bytes(),
-                    Bundle.from_json(asset("tsa/bundle.txt.sigstore").read_bytes()),
-                    null_policy,
-                )
-
-        assert (
-            "Unable to verify Timestamp because no validity provided."
-            == caplog.records[0].message
-        )
-
     def test_verifier_outside_validity_range(
         self, caplog, verifier, asset, null_policy, monkeypatch
     ):
@@ -356,11 +386,6 @@ class TestVerifierWithTimestamp:
                 null_policy,
             )
 
-        assert (
-            "Error while verifying certificates: Unable to verify pkcs7 signature"
-            in caplog.records[0].message
-        )
-
     def test_verifier_not_enough_timestamp(
         self, verifier, asset, null_policy, monkeypatch
     ):
@@ -375,3 +400,18 @@ class TestVerifierWithTimestamp:
                 Bundle.from_json(asset("tsa/bundle.txt.sigstore").read_bytes()),
                 null_policy,
             )
+
+    def test_verify_signed_timestamp_regression(self, asset):
+        """
+        Ensure we correctly verify a timestamp with no embedded certs.
+
+        This is a regression test for # 1482
+        """
+        verifier = Verifier.staging(offline=True)
+        ts = rfc3161_client.decode_timestamp_response(
+            asset("tsa/issue1482-timestamp-with-no-cert").read_bytes()
+        )
+        res = verifier._verify_signed_timestamp(
+            ts, asset("tsa/issue1482-message").read_bytes()
+        )
+        assert res is not None
